//
// Copyright 2016 frankmuenster.de
// 
// Lizenziert unter der EUPL, Version 1.1 oder - sobald
// diese von der Europischen Kommission genehmigt wurden -
// Folgeversionen der EUPL ("Lizenz");
// Sie drfen dieses Werk ausschlielich gem
// dieser Lizenz nutzen.
// Eine Kopie der Lizenz finden Sie hier:
// 
//   http://ec.europa.eu/idabc/eupl
// 
// Sofern nicht durch anwendbare Rechtsvorschriften
// gefordert oder in schriftlicher Form vereinbart, wird
// die unter der Lizenz verbreitete Software "so wie sie
// ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
// ausdrcklich oder stillschweigend - verbreitet.
// Die sprachspezifischen Genehmigungen und Beschrnkungen
// unter der Lizenz sind dem Lizenztext zu entnehmen.
// 

package de.frankmuenster.hibiscus.mahoe;

import java.rmi.RemoteException;

import de.frankmuenster.hibiscus.mahoe.messages.AbstractBankMessages;
import de.frankmuenster.hibiscus.mahoe.utils.I18N;
import de.frankmuenster.hibiscus.mahoe.web.AbstractWebSite;
import de.frankmuenster.hibiscus.mahoe.web.WebAccountInformation;
import de.frankmuenster.hibiscus.mahoe.web.WebSiteException;
import de.willuhn.jameica.hbci.rmi.Konto;
import de.willuhn.jameica.hbci.synchronize.jobs.SynchronizeJobKontoauszug;
import de.willuhn.logging.Logger;
import de.willuhn.util.ApplicationException;
import de.willuhn.util.ProgressMonitor;

/**
 * Abstrakte Klasse fr Kontoauszugsabfrage fr Mahoe Konten. Hier verarbeiten
 * wir hauptschlich die Fehlerbehandlung.
 * 
 * @author Frank Mnster
 * 
 * @param <I>
 *            die bankspezifischen WebAccountInformation
 * @param <M>
 *            die bankspezifische AbstractBankMessages
 * 
 */
public abstract class AbstractMahoeSynchronizeJobKontoauszug<I extends WebAccountInformation, M extends AbstractBankMessages>
		extends SynchronizeJobKontoauszug implements InterfaceMahoeSynchronizeJob {

	/**
	 * Das Konto
	 */
	protected transient Konto konto;

	/**
	 * Die Kundennummer. Extra damit man nicht jedes mal die RemoteException
	 * abfangen muss.
	 */
	protected transient String kundennummer;

	/**
	 * Das Webseiten Objekt
	 */
	protected transient AbstractWebSite<I, M> webSite;

	/**
	 * Kontoinformationen
	 */
	protected transient I accountInfo;

	/**
	 * Das Progessmonitor Objekt
	 */
	protected transient ProgressMonitor monitor;

	/**
	 * Standardkonstruktor
	 */
	public AbstractMahoeSynchronizeJobKontoauszug() {
		super();
	}

	/**
	 * @see InterfaceMahoeSynchronizeJob#execute(ProgressMonitor, boolean)
	 */
	@Override
	public void execute(final ProgressMonitor monitor, final boolean lastJob) throws Exception { // NOPMD

		this.konto = (Konto) this.getContext(CTX_ENTITY);
		this.monitor = monitor;

		// Starte Abfrage
		final Konto konto = (Konto) this.getContext(CTX_ENTITY);

		String accountname;
		try {
			kundennummer = konto.getKundennummer();
			accountname = konto.getBezeichnung();
		} catch (RemoteException e) {
			throw new ApplicationException(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.readAccountInfoFailed"), e); //$NON-NLS-1$
		}
		Logger.info(
				I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.readTransactionLoginfo", accountname, kundennummer)); //$NON-NLS-1$

		// Wir laden das BankSeiten-Objekt
		getWebSite();
		monitor.setPercentComplete(10);

		// Nur Anmelden, wenn die Verbindung nicht schon besteht
		if (webSite.isStillLoggedIn()) {
			monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.connectionStillValid")); //$NON-NLS-1$
		} else {
			monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.startLogin")); //$NON-NLS-1$
			login();
			Logger.info(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.loginSuccessful")); //$NON-NLS-1$
			monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.loginSuccessful")); //$NON-NLS-1$
		}
		monitor.setPercentComplete(20);

		// Kontoinformationen
		monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.updateAccountInfo")); //$NON-NLS-1$
		accountInfo = webSite.getAccountInformation();
		accountInfo.saveInformationInAccount(monitor);
		monitor.setPercentComplete(30);

		// Umstze
		monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.readTransactions")); //$NON-NLS-1$
		handleTransactions();
		monitor.setPercentComplete(60);

		// Kontoauszge
		monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.readAccountSheets")); //$NON-NLS-1$
		handleAccountSheets();
		monitor.setPercentComplete(60);

		// Banknachrichten
		try {
			monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.readMessages")); //$NON-NLS-1$
			final int savedMsg = webSite.getMessages(accountInfo);
			if (savedMsg == 0) {
				monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.noNewMessages")); //$NON-NLS-1$
			} else {
				monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.newMessagesStored")); //$NON-NLS-1$
			}
		} catch (WebSiteException | ApplicationException e) {
			monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.readMessagesFailed")); //$NON-NLS-1$
			Logger.error(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.readMessagesFailedLog", e.getMessage()), e); //$NON-NLS-1$
		}

		// Falls letzter Job fr die Bank, abmelden
		if (lastJob) {
			monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.reachedLastJob")); //$NON-NLS-1$
			webSite.logoff();
		}
		monitor.setPercentComplete(100);

	}

	/**
	 * Liefert das Webseiten-Objekt fr die Bank.
	 * 
	 * @throws WebSiteException
	 * @throws ApplicationException
	 */
	protected abstract void getWebSite() throws WebSiteException, ApplicationException;

	/**
	 * Meldet das Webseiten-Objekt an der Bank an.
	 * 
	 * @throws WebSiteException
	 * @throws ApplicationException
	 */
	protected abstract void login() throws WebSiteException, ApplicationException;

	/**
	 * Liest und speichert die Umstze. Ob auch vorgemerkte Umstze bearbeitet
	 * werden, entscheidet die konkrete Implementierung.
	 * 
	 * @throws WebSiteException
	 * @throws ApplicationException
	 */
	protected abstract void handleTransactions() throws WebSiteException, ApplicationException;

	/**
	 * Speichert die Kontoauszge als PDF.
	 * 
	 * @throws WebSiteException
	 * @throws ApplicationException
	 */
	protected abstract void handleAccountSheets() throws WebSiteException, ApplicationException;
}
