//
// Copyright 2016 frankmuenster.de
//
// Lizenziert unter der EUPL, Version 1.1 oder - sobald
// diese von der Europischen Kommission genehmigt wurden -
// Folgeversionen der EUPL ("Lizenz");
// Sie drfen dieses Werk ausschlielich gem
// dieser Lizenz nutzen.
// Eine Kopie der Lizenz finden Sie hier:
//
//   http://ec.europa.eu/idabc/eupl
//
// Sofern nicht durch anwendbare Rechtsvorschriften
// gefordert oder in schriftlicher Form vereinbart, wird
// die unter der Lizenz verbreitete Software "so wie sie
// ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
// ausdrcklich oder stillschweigend - verbreitet.
// Die sprachspezifischen Genehmigungen und Beschrnkungen
// unter der Lizenz sind dem Lizenztext zu entnehmen.
//

package de.frankmuenster.hibiscus.mahoe.santander.pdfextractor;

import java.io.IOException;
import java.io.InputStream;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.text.NumberFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.List;
import java.util.Locale;

import org.apache.commons.lang.StringUtils;
import org.apache.pdfbox.pdmodel.PDPage;

import de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractBankPDFPage;
import de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractPdfExtractor;
import de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractTransactionCreator;
import de.frankmuenster.hibiscus.mahoe.pdfextractor.PdfBuchung;
import de.frankmuenster.hibiscus.mahoe.pdfextractor.PdfEintrag;
import de.frankmuenster.hibiscus.mahoe.santander.clearing.ClearingEntry;
import de.frankmuenster.hibiscus.mahoe.utils.I18N;
import de.frankmuenster.hibiscus.mahoe.utils.Utils;
import de.willuhn.logging.Logger;

/**
 * Class for extracting entry information from a TargoBank PDF bank statement.
 *
 * @author Frank Mnster
 *
 */
public class SantanderPdfExtractor extends AbstractPdfExtractor {

	/**
	 * Zeilenhhe der Buchungszeile im MahoeMain PDF.
	 */
	private static final double LINE_HEIGH = 9.35;

	/**
	 *
	 */
	private static final double MAX_POS_DATE_DIFF = 0.05;

	/**
	 * Dezimalformatierer
	 */
	private final static NumberFormat N_FORMAT = NumberFormat.getInstance(Locale.GERMAN);

	/**
	 * Liste mit den Fremdw&auml;hrungen
	 */
	private transient List<PdfEintrag> fremdwaehrung;

	/**
	 * Liste mit den Fremdw&auml;hrungsbetr&auml;gen
	 */
	private transient List<PdfEintrag> fwBetraege;

	/**
	 * Unser TransactionCreator
	 */
	private transient SantanderTransactionCreator transCreator;

	/**
	 * Erstellt einen PDF-Extraktor der Santander Bank
	 *
	 * @see AbstractPdfExtractor#AbstractPdfExtractor(InputStream, String)
	 *
	 */
	public SantanderPdfExtractor(final InputStream inputStream, final String kartenNummer) {
		super(inputStream, kartenNummer);
	}

	/**
	 * Liefert den Ausgleichbetrag. Das ist die Differenz zwischen dem Saldo und
	 * dem flligem Betrag.
	 *
	 * @return den Ausgleichbetrag
	 */
	public ClearingEntry getAusgleichsBuchung() {

		ClearingEntry buchung = null;
		final PDPage page = this.document.getPage(0);
		// Logger.debug("Bearbeite Seite: " + pageNumber);
		try {
			final AbstractBankPDFPage textAreas = new SantanderPdfFirstPage();
			textAreas.extractRegions(page);

			final String kartenkonto = this.getKartenkonto(textAreas);
			final String karteninhaber = super.getKarteninhaber(textAreas);
			final String iban = super.getIban(textAreas);
			final String bic = super.getBic(textAreas);
			final BigDecimal saldo = new BigDecimal(this.getSaldo(textAreas) * -1).setScale(2, RoundingMode.HALF_UP);
			final BigDecimal fBetrag = this.getFaelligBetrag(textAreas).setScale(2, RoundingMode.HALF_UP);
			final BigDecimal betrag = saldo.subtract(fBetrag).setScale(2, RoundingMode.HALF_UP);
			final SimpleDateFormat sdf = new SimpleDateFormat("dd.MM.yyyy", Locale.GERMAN); //$NON-NLS-1$
			final Date datum = sdf.parse(this.getFaelligDatum(textAreas));

			buchung = new ClearingEntry(kartenkonto, karteninhaber, iban, bic, datum, betrag);

		} catch (IOException | ParseException e) {
			Logger.warn(I18N.tr("SantanderPdfExtractor.extractingDataFailed") + e.getMessage()); //$NON-NLS-1$
			Logger.warn(e.getStackTrace().toString());
		}

		return buchung;

	}

	/**
	 * Liefert das Flligkeitsdatum des Saldos
	 *
	 * @param textAreas
	 *            die Textares der ersten Seite
	 *
	 * @return das Flligkeitsdatum
	 */
	private String getFaelligDatum(final AbstractBankPDFPage textAreas) {

		final String[] parsedText = textAreas.getTextForRegion(AbstractBankPDFPage.FAELLIG_DATUM)
				.split(AbstractPdfExtractor.WORT_DELIMITER);
		return Utils.removeEOL(parsedText[2]);
	}

	private BigDecimal getFaelligBetrag(final AbstractBankPDFPage textAreas) {

		final String[] parsedText = textAreas.getTextForRegion(AbstractBankPDFPage.FAELLIG_BETRAG)
				.split(AbstractPdfExtractor.WORT_DELIMITER);
		// Logger.debug("PT (kartenkonto): " + parsedText.length);
		String fBetragString = Utils.removeEOL(parsedText[2]);
		fBetragString = fBetragString.replace(" ", StringUtils.EMPTY); //$NON-NLS-1$

		BigDecimal fBetrag = null;
		try {
			fBetrag = new BigDecimal(this.numFormat.parse(fBetragString).doubleValue());
		} catch (ParseException e) {
			Logger.warn(I18N.tr("SantanderPdfExtractor.clearingAmountNotFound") + e.getMessage()); //$NON-NLS-1$
		}
		return fBetrag;
	}

	private void getFremdWaehrungen(final AbstractBankPDFPage textAreas) {

		this.fremdwaehrung = new ArrayList<>();
		this.fwBetraege = new ArrayList<>();

		String regioText = textAreas.getTextForRegion(AbstractBankPDFPage.FREMDW_BETRAG);
		final String lineSeparator = textAreas.getLineSeparator();

		if (lineSeparator.equals(regioText)) {
			regioText = StringUtils.EMPTY;
		}

		final String[] lines = regioText.split(lineSeparator);

		boolean multiLine = false;
		String betragText = StringUtils.EMPTY;
		String waehrungText = StringUtils.EMPTY;

		for (final String line : lines) {
			Logger.debug("Line (fwBetrag): >" + line + "<"); //$NON-NLS-1$//$NON-NLS-2$

			if (StringUtils.isNotEmpty(line)) {
				final String[] woerter = line.split(" "); //$NON-NLS-1$

				Logger.debug("Anzahl Wrter: " + woerter.length); //$NON-NLS-1$
				if (woerter.length == AbstractPdfExtractor.MAX_ONE_WORD) {
					if (multiLine) {
						multiLine = false;
						betragText = woerter[0];
					} else {
						multiLine = true;
						waehrungText = woerter[0];
						continue;
					}
				} else if (woerter.length == AbstractPdfExtractor.MAX_TWO_WORD) {
					multiLine = false;
					waehrungText = woerter[0];
					betragText = woerter[1];
				} else {
					break;
				}

				final PdfEintrag pdfWaehrung = new PdfEintrag(); // NOPMD
				if (StringUtils.isEmpty(waehrungText)) {
					pdfWaehrung.setWort(waehrungText);
				} else {
					final String[] wortInfo = waehrungText.split(AbstractPdfExtractor.WORT_DELIMITER);
					pdfWaehrung.setPositionX(Double.valueOf(wortInfo[0]));
					pdfWaehrung.setPositionY(Double.valueOf(wortInfo[1]));
					pdfWaehrung.setWort(Utils.removeEOL(wortInfo[2]));
				}

				final PdfEintrag pdfBetrag = new PdfEintrag(); // NOPMD
				if (StringUtils.isEmpty(betragText)) {
					pdfBetrag.setWort(betragText);
				} else {
					final String[] wortInfo = betragText.split(AbstractPdfExtractor.WORT_DELIMITER);
					pdfBetrag.setPositionX(Double.valueOf(wortInfo[0]));
					pdfBetrag.setPositionY(pdfWaehrung.getPositionY());
					pdfBetrag.setWort(Utils.removeEOL(wortInfo[2]));
				}

				this.fremdwaehrung.add(pdfWaehrung);
				this.fwBetraege.add(pdfBetrag);
			}
		}

		// Anmerkungen entfernen
		this.removeAnnotations(this.fremdwaehrung);
		this.removeAnnotations(this.fwBetraege);

	}

	/**
	 * @see de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractPdfExtractor#addYear
	 */
	@Override
	protected void addYear(final List<PdfEintrag> entries) {

		// Jahr hinzu tun
		final int rMonat = Integer.valueOf(this.rechungsDatum.substring(3, 5));
		final int rJahr = Integer.valueOf(this.rechungsDatum.substring(6, 10));

		for (final PdfEintrag entry : entries) {
			final String datum = entry.getWort().substring(0, 2) + "." //$NON-NLS-1$
					+ entry.getWort().substring(2, 4) + "."; //$NON-NLS-1$
			final int bMonat = Integer.valueOf(entry.getWort().substring(2, 4));
			if (bMonat > rMonat) {
				entry.setWort(datum + (rJahr - 1));
			} else {
				entry.setWort(datum + rJahr);
			}
		}
	}

	/**
	 * @throws IOException
	 * @see de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractPdfExtractor#getBankPdfEvenPage()
	 */
	@Override
	protected AbstractBankPDFPage getBankPdfEvenPage() throws IOException {
		return new SantanderPdfOtherPage();
	}

	/**
	 * @throws IOException
	 * @see de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractPdfExtractor#getBankPdfFirstPage()
	 */
	@Override
	protected AbstractBankPDFPage getBankPdfFirstPage() throws IOException {
		return new SantanderPdfFirstPage();
	}

	/**
	 * @throws IOException
	 * @see de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractPdfExtractor#getBankPdfOddPage()
	 */
	@Override
	protected AbstractBankPDFPage getBankPdfOddPage() throws IOException {
		return new SantanderPdfOtherPage();
	}

	/**
	 * @see de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractPdfExtractor#getBetraege(de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractBankPDFPage)
	 */
	@Override
	protected List<PdfEintrag> getBetraege(final AbstractBankPDFPage textAreas) {

		final List<PdfEintrag> betraege = new ArrayList<>();

		final String regioText = textAreas.getTextForRegion(AbstractBankPDFPage.BETRAG);
		final String lineSeparator = textAreas.getLineSeparator();
		final String[] lines = regioText.split(lineSeparator);

		for (final String line : lines) {
			Logger.trace("Line (betrag): >" + line + "<"); //$NON-NLS-1$ //$NON-NLS-2$

			if (StringUtils.isNotEmpty(line)) {
				final String[] woerter = line.split(" "); //$NON-NLS-1$

				if (woerter.length == 0 || woerter.length > AbstractPdfExtractor.MAX_TWO_WORD) {
					break;
				}
				boolean isHaben = false;
				if (woerter.length == AbstractPdfExtractor.MAX_TWO_WORD) {
					// Wenn das zweite Wort nur ein H ist, haben wir eine
					// Einzahlung
					final String[] testHaben = woerter[1].split(AbstractPdfExtractor.WORT_DELIMITER);
					if ("H".equals(testHaben[2].trim())) { //$NON-NLS-1$
						isHaben = true;
					}
				}
				final String wort = woerter[0];
				final String[] wortInfo = wort.split(AbstractPdfExtractor.WORT_DELIMITER);
				if (wortInfo.length != AbstractPdfExtractor.WORT_INFO_LENGTH) {
					continue;
				}
				if (!isHaben) {
					wortInfo[2] = '-' + wortInfo[2]; // NOPMD
				}
				final PdfEintrag pdfEintrag = new PdfEintrag(); // NOPMD
				pdfEintrag.setPositionX(Double.valueOf(wortInfo[0]));
				pdfEintrag.setPositionY(Double.valueOf(wortInfo[1]));
				pdfEintrag.setWort(Utils.removeEOL(wortInfo[2]));
				betraege.add(pdfEintrag);
			}
		}

		// Anmerkungen entfernen
		this.removeAnnotations(betraege);

		return betraege;
	}

	/**
	 * @see AbstractPdfExtractor#getFremdwaehrungen(AbstractBankPDFPage)
	 */
	@Override
	protected List<PdfEintrag> getFremdwaehrungen(final AbstractBankPDFPage textAreas) {

		if (this.fremdwaehrung == null) {
			this.getFremdWaehrungen(textAreas);
		}
		return this.fremdwaehrung;

	}

	/**
	 * @see de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractPdfExtractor#getFremdwaehrungsBetraege(de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractBankPDFPage)
	 */
	@Override
	protected List<PdfEintrag> getFremdwaehrungsBetraege(final AbstractBankPDFPage textAreas) {

		if (this.fwBetraege == null) {
			this.getFremdWaehrungen(textAreas);
		}
		return this.fwBetraege;
	}

	/**
	 * @see de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractPdfExtractor#getNumberFormat()
	 */
	@Override
	protected NumberFormat getNumberFormat() {
		return SantanderPdfExtractor.N_FORMAT;
	}

	/**
	 * @see de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractPdfExtractor#getTransactionCreator()
	 */
	@Override
	protected AbstractTransactionCreator getTransactionCreator() {
		if (this.transCreator == null) {
			this.transCreator = new SantanderTransactionCreator(this.buchungsDaten, SantanderPdfExtractor.LINE_HEIGH);
		}
		return this.transCreator;
	}

	/**
	 * @see de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractPdfExtractor#removeAnnotations(java.util.List)
	 */
	@Override
	protected void removeAnnotations(final List<PdfEintrag> pdfWortList) {
		final double yOfFirstDatum = this.buchungsDaten.get(this.buchungsDaten.size() - 1).getPositionY() - 10;
		Logger.trace("First: " + yOfFirstDatum); //$NON-NLS-1$
		for (int i = pdfWortList.size() - 1; i >= 0; i--) {
			Logger.trace("Check: " + pdfWortList.get(i).getPositionY()); //$NON-NLS-1$

			if (pdfWortList.get(i).getPositionY() < yOfFirstDatum) {
				Logger.trace("Remove"); //$NON-NLS-1$
				pdfWortList.remove(i);
			}
		}
	}

	/**
	 * @see de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractPdfExtractor#removeAnnotationsBookingDates(java.util.List)
	 */
	@Override
	protected void removeAnnotationsBookingDates(final List<PdfEintrag> buchungenDatum) {

		if (!buchungenDatum.isEmpty()) {
			final double xOfFirstDatum = buchungenDatum.get(0).getPositionX();
			Logger.trace("First date: " + xOfFirstDatum); //$NON-NLS-1$
			for (int i = buchungenDatum.size() - 1; i > 0; i--) {
				Logger.trace("Check date x position: " + buchungenDatum.get(i).getPositionX()); //$NON-NLS-1$
				final double xPosDiff = Math.abs(buchungenDatum.get(i).getPositionX() - xOfFirstDatum);
				if (xPosDiff > SantanderPdfExtractor.MAX_POS_DATE_DIFF) {
					Logger.trace("Remove date as it is an annotation in the bottom"); //$NON-NLS-1$
					buchungenDatum.remove(i);
				}
			}
		}
	}

	/**
	 * @see de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractPdfExtractor#reOrder(java.util.List)
	 */
	@Override
	protected void reOrder(final List<PdfBuchung> buchungen) {

		// Leider mssen wir richtig sortieren, da die Umstze nach Karten
		// sortiert sind
		Collections.sort(buchungen, new Comparator<PdfBuchung>() {

			/**
			 * @see Comparator#compare
			 *
			 */
			@Override
			public int compare(final PdfBuchung buchung1, final PdfBuchung buchung2) {
				return (buchung1.getBuchungsDatum().compareTo(buchung2.getBuchungsDatum())) * -1;
			}
		});

	}

}
